@@ -15,10 +15,12 @@ import logging
 from typing import Any, Callable, Optional
 
 from awscrt import mqtt
+from awscrt.exceptions import AwsCrtError
 
 from .events import EventEmitter
+from .exceptions import MqttNotConnectedError
 from .models import Device, DeviceFeature, DeviceStatus, EnergyUsageResponse
-from .mqtt_utils import redact_topic
+from .mqtt_utils import redact_topic, topic_matches_pattern
 
 __author__ = "Emmanuel Levijarvi"
 
@@ -60,7 +62,9 @@ class MqttSubscriptionManager:
 
         # Track subscriptions and handlers
         self._subscriptions: dict[str, mqtt.QoS] = {}
-        self._message_handlers: dict[str, list[Callable[[str, dict[str, Any]], None]]] = {}
+        self._message_handlers: dict[
+            str, list[Callable[[str, dict[str, Any]], None]]
+        ] = {}
 
         # Track previous state for change detection
         self._previous_status: Optional[DeviceStatus] = None
@@ -70,9 +74,28 @@ class MqttSubscriptionManager:
         """Get current subscriptions."""
         return self._subscriptions.copy()
 
-    def _on_message_received(self, topic: str, payload: bytes, **kwargs: Any) -> None:
+    def update_connection(self, connection: Any) -> None:
         """
-        Internal callback for received messages.
+        Update the MQTT connection reference.
+
+        This is used when the connection is recreated (e.g., after reconnection)
+        to update the internal reference while preserving subscriptions.
+
+        Args:
+            connection: New MQTT connection object
+
+        Note:
+            This does not re-establish subscriptions. Call the appropriate
+            subscribe methods to re-register subscriptions with the new
+            connection if needed.
+        """
+        self._connection = connection
+        _logger.debug("Updated subscription manager connection reference")
+
+    def _on_message_received(
+        self, topic: str, payload: bytes, **kwargs: Any
+    ) -> None:
+        """Handle received MQTT messages.
 
         Parses JSON payload and routes to registered handlers.
 
@@ -88,73 +111,22 @@ class MqttSubscriptionManager:
 
             # Call registered handlers that match this topic
             # Need to match against subscription patterns with wildcards
-            for subscription_pattern, handlers in self._message_handlers.items():
-                if self._topic_matches_pattern(topic, subscription_pattern):
+            for (
+                subscription_pattern,
+                handlers,
+            ) in self._message_handlers.items():
+                if topic_matches_pattern(topic, subscription_pattern):
                     for handler in handlers:
                         try:
                             handler(topic, message)
-                        except Exception as e:
+                        except (TypeError, AttributeError, KeyError) as e:
                             _logger.error(f"Error in message handler: {e}")
 
         except json.JSONDecodeError as e:
             _logger.error(f"Failed to parse message payload: {e}")
-        except Exception as e:
+        except (AttributeError, KeyError, TypeError) as e:
             _logger.error(f"Error processing message: {e}")
 
-    def _topic_matches_pattern(self, topic: str, pattern: str) -> bool:
-        """
-        Check if a topic matches a subscription pattern with wildcards.
-
-        Supports MQTT wildcards:
-        - '+' matches a single level
-        - '#' matches multiple levels (must be at end)
-
-        Args:
-            topic: Actual topic (e.g., "cmd/52/navilink-ABC/status")
-            pattern: Pattern with wildcards (e.g., "cmd/52/+/#")
-
-        Returns:
-            True if topic matches pattern
-
-        Examples:
-            >>> _topic_matches_pattern("cmd/52/device1/status", "cmd/52/+/status")
-            True
-            >>> _topic_matches_pattern("cmd/52/device1/status/extra", "cmd/52/device1/#")
-            True
-        """
-        # Handle exact match
-        if topic == pattern:
-            return True
-
-        # Handle wildcards
-        topic_parts = topic.split("/")
-        pattern_parts = pattern.split("/")
-
-        # Multi-level wildcard # matches everything after
-        if "#" in pattern_parts:
-            hash_idx = pattern_parts.index("#")
-            # Must be at the end
-            if hash_idx != len(pattern_parts) - 1:
-                return False
-            # Topic must have at least as many parts as before the #
-            if len(topic_parts) < hash_idx:
-                return False
-            # Check parts before # with + wildcard support
-            for i in range(hash_idx):
-                if pattern_parts[i] != "+" and topic_parts[i] != pattern_parts[i]:
-                    return False
-            return True
-
-        # Single-level wildcard + matches one level
-        if len(topic_parts) != len(pattern_parts):
-            return False
-
-        for topic_part, pattern_part in zip(topic_parts, pattern_parts):
-            if pattern_part != "+" and topic_part != pattern_part:
-                return False
-
-        return True
-
     async def subscribe(
         self,
         topic: str,
@@ -177,19 +149,34 @@ class MqttSubscriptionManager:
             Exception: If subscription fails
         """
         if not self._connection:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         _logger.info(f"Subscribing to topic: {redact_topic(topic)}")
 
         try:
             # Convert concurrent.futures.Future to asyncio.Future and await
+            # Use shield to prevent cancellation from propagating to
+            # underlying future
             subscribe_future, packet_id = self._connection.subscribe(
                 topic=topic, qos=qos, callback=self._on_message_received
             )
-            subscribe_result = await asyncio.wrap_future(subscribe_future)
+            try:
+                subscribe_result = await asyncio.shield(
+                    asyncio.wrap_future(subscribe_future)
+                )
+            except asyncio.CancelledError:
+                # Shield was cancelled - the underlying subscribe will
+                # complete independently, preventing InvalidStateError
+                # in AWS CRT callbacks
+                _logger.debug(
+                    f"Subscribe to '{redact_topic(topic)}' was cancelled "
+                    "but will complete in background"
+                )
+                raise
 
             _logger.info(
-                f"Subscribed to '{redact_topic(topic)}' with QoS {subscribe_result['qos']}"
+                f"Subscription succeeded (topic redacted) with QoS "
+                f"{subscribe_result['qos']}"
             )
 
             # Store subscription and handler
@@ -200,8 +187,10 @@ class MqttSubscriptionManager:
 
             return int(packet_id)
 
-        except Exception as e:
-            _logger.error(f"Failed to subscribe to '{redact_topic(topic)}': {e}")
+        except (AwsCrtError, RuntimeError) as e:
+            _logger.error(
+                f"Failed to subscribe to '{redact_topic(topic)}': {e}"
+            )
             raise
 
     async def unsubscribe(self, topic: str) -> int:
@@ -219,14 +208,26 @@ class MqttSubscriptionManager:
             Exception: If unsubscribe fails
         """
         if not self._connection:
-            raise RuntimeError("Not connected to MQTT broker")
+            raise MqttNotConnectedError("Not connected to MQTT broker")
 
         _logger.info(f"Unsubscribing from topic: {redact_topic(topic)}")
 
         try:
             # Convert concurrent.futures.Future to asyncio.Future and await
+            # Use shield to prevent cancellation from propagating to
+            # underlying future
             unsubscribe_future, packet_id = self._connection.unsubscribe(topic)
-            await asyncio.wrap_future(unsubscribe_future)
+            try:
+                await asyncio.shield(asyncio.wrap_future(unsubscribe_future))
+            except asyncio.CancelledError:
+                # Shield was cancelled - the underlying unsubscribe will
+                # complete independently, preventing InvalidStateError
+                # in AWS CRT callbacks
+                _logger.debug(
+                    f"Unsubscribe from '{redact_topic(topic)}' was "
+                    "cancelled but will complete in background"
+                )
+                raise
 
             # Remove from tracking
             self._subscriptions.pop(topic, None)
@@ -236,10 +237,75 @@ class MqttSubscriptionManager:
 
             return int(packet_id)
 
-        except Exception as e:
-            _logger.error(f"Failed to unsubscribe from '{redact_topic(topic)}': {e}")
+        except (AwsCrtError, RuntimeError) as e:
+            _logger.error(
+                f"Failed to unsubscribe from '{redact_topic(topic)}': {e}"
+            )
             raise
 
+    async def resubscribe_all(self) -> None:
+        """
+        Re-establish all subscriptions after a connection rebuild.
+
+        This method is called after a deep reconnection to restore all
+        active subscriptions. It uses the stored subscription information
+        to re-subscribe to all topics with their original QoS settings
+        and handlers.
+
+        Note:
+            This is typically called automatically during deep reconnection
+            and should not need to be called manually.
+
+        Raises:
+            RuntimeError: If not connected to MQTT broker
+            Exception: If any subscription fails
+        """
+        if not self._connection:
+            raise MqttNotConnectedError("Not connected to MQTT broker")
+
+        if not self._subscriptions:
+            _logger.debug("No subscriptions to restore")
+            return
+
+        subscription_count = len(self._subscriptions)
+        _logger.info(f"Re-establishing {subscription_count} subscription(s)...")
+
+        # Store subscriptions to re-establish (avoid modifying dict during
+        # iteration)
+        subscriptions_to_restore = list(self._subscriptions.items())
+        handlers_to_restore = {
+            topic: handlers.copy()
+            for topic, handlers in self._message_handlers.items()
+        }
+
+        # Clear current subscriptions (will be re-added by subscribe())
+        self._subscriptions.clear()
+        self._message_handlers.clear()
+
+        # Re-establish each subscription
+        failed_subscriptions = set()
+        for topic, qos in subscriptions_to_restore:
+            handlers = handlers_to_restore.get(topic, [])
+            for handler in handlers:
+                try:
+                    await self.subscribe(topic, handler, qos)
+                except (AwsCrtError, RuntimeError) as e:
+                    _logger.error(
+                        f"Failed to re-subscribe to "
+                        f"'{redact_topic(topic)}': {e}"
+                    )
+                    # Mark topic as failed and skip remaining handlers
+                    # since they will fail for the same reason
+                    failed_subscriptions.add(topic)
+                    break  # Exit handler loop, move to next topic
+
+        if failed_subscriptions:
+            _logger.warning(
+                f"Failed to restore {len(failed_subscriptions)} subscription(s)"
+            )
+        else:
+            _logger.info("All subscriptions re-established successfully")
+
     async def subscribe_device(
         self, device: Device, callback: Callable[[str, dict[str, Any]], None]
     ) -> int:
@@ -291,8 +357,8 @@ class MqttSubscriptionManager:
         Example (Traditional Callback)::
 
             >>> def on_status(status: DeviceStatus):
-            ...     print(f"Temperature: {status.dhwTemperature}°F")
-            ...     print(f"Mode: {status.operationMode}")
+            ...     print(f"Temperature: {status.dhw_temperature}°F")
+            ...     print(f"Mode: {status.operation_mode}")
             >>>
             >>> await mqtt_client.subscribe_device_status(device, on_status)
 
@@ -304,23 +370,27 @@ class MqttSubscriptionManager:
             >>>
             >>> # State change events
             >>> mqtt_client.on('heating_started', lambda s: print("Heating ON"))
-            >>> mqtt_client.on('heating_stopped', lambda s: print("Heating OFF"))
+            >>> mqtt_client.on('heating_stopped', lambda s: print("Heating
+            OFF"))
             >>>
             >>> # Subscribe to start receiving events
-            >>> await mqtt_client.subscribe_device_status(device, lambda s: None)
+            >>> await mqtt_client.subscribe_device_status(device, lambda s:
+            None)
         """
 
         def status_message_handler(topic: str, message: dict[str, Any]) -> None:
             """Parse status messages and invoke user callback."""
             try:
                 # Log all messages received for debugging
-                _logger.debug(f"Status handler received message on topic: {topic}")
+                _logger.debug(
+                    f"Status handler received message on topic: {topic}"
+                )
                 _logger.debug(f"Message keys: {list(message.keys())}")
 
-                # Check if message contains status data
                 if "response" not in message:
                     _logger.debug(
-                        "Message does not contain 'response' key, skipping. Keys: %s",
+                        "Message does not contain 'response' key, skipping. "
+                        "Keys: %s",
                         list(message.keys()),
                     )
                     return
@@ -330,21 +400,28 @@ class MqttSubscriptionManager:
 
                 if "status" not in response:
                     _logger.debug(
-                        "Response does not contain 'status' key, skipping. Keys: %s",
+                        "Response does not contain 'status' key, skipping. "
+                        "Keys: %s",
                         list(response.keys()),
                     )
                     return
 
                 # Parse status into DeviceStatus object
-                _logger.info(f"Parsing device status message from topic: {topic}")
+                _logger.info(
+                    f"Parsing device status message from topic: {topic}"
+                )
                 status_data = response["status"]
                 device_status = DeviceStatus.from_dict(status_data)
 
                 # Emit raw status event
-                self._schedule_coroutine(self._event_emitter.emit("status_received", device_status))
+                self._schedule_coroutine(
+                    self._event_emitter.emit("status_received", device_status)
+                )
 
                 # Detect and emit state changes
-                self._schedule_coroutine(self._detect_state_changes(device_status))
+                self._schedule_coroutine(
+                    self._detect_state_changes(device_status)
+                )
 
                 # Invoke user callback with parsed status
                 _logger.info("Invoking user callback with parsed DeviceStatus")
@@ -357,12 +434,18 @@ class MqttSubscriptionManager:
                     exc_info=True,
                 )
             except ValueError as e:
-                _logger.warning(f"Invalid value in status message: {e}", exc_info=True)
-            except Exception as e:
-                _logger.error(f"Error parsing device status: {e}", exc_info=True)
+                _logger.warning(
+                    f"Invalid value in status message: {e}", exc_info=True
+                )
+            except (TypeError, AttributeError) as e:
+                _logger.error(
+                    f"Error parsing device status: {e}", exc_info=True
+                )
 
         # Subscribe using the internal handler
-        return await self.subscribe_device(device=device, callback=status_message_handler)
+        return await self.subscribe_device(
+            device=device, callback=status_message_handler
+        )
 
     async def _detect_state_changes(self, status: DeviceStatus) -> None:
         """
@@ -383,39 +466,44 @@ class MqttSubscriptionManager:
 
         try:
             # Temperature change
-            if status.dhwTemperature != prev.dhwTemperature:
+            if status.dhw_temperature != prev.dhw_temperature:
                 await self._event_emitter.emit(
                     "temperature_changed",
-                    prev.dhwTemperature,
-                    status.dhwTemperature,
+                    prev.dhw_temperature,
+                    status.dhw_temperature,
                 )
                 _logger.debug(
-                    f"Temperature changed: {prev.dhwTemperature}°F → {status.dhwTemperature}°F"
+                    f"Temperature changed: {prev.dhw_temperature}°F → "
+                    f"{status.dhw_temperature}°F"
                 )
 
             # Operation mode change
-            if status.operationMode != prev.operationMode:
+            if status.operation_mode != prev.operation_mode:
                 await self._event_emitter.emit(
                     "mode_changed",
-                    prev.operationMode,
-                    status.operationMode,
-                )
-                _logger.debug(f"Mode changed: {prev.operationMode} → {status.operationMode}")
-
-            # Power consumption change
-            if status.currentInstPower != prev.currentInstPower:
-                await self._event_emitter.emit(
-                    "power_changed",
-                    prev.currentInstPower,
-                    status.currentInstPower,
+                    prev.operation_mode,
+                    status.operation_mode,
                 )
                 _logger.debug(
-                    f"Power changed: {prev.currentInstPower}W → {status.currentInstPower}W"
+                    f"Mode changed: {prev.operation_mode} → "
+                    f"{status.operation_mode}"
+                )
+
+            # Power consumption change
+            if status.current_inst_power != prev.current_inst_power:
+                await self._event_emitter.emit(
+                    "power_changed",
+                    prev.current_inst_power,
+                    status.current_inst_power,
+                )
+                _logger.debug(
+                    f"Power changed: {prev.current_inst_power}W → "
+                    f"{status.current_inst_power}W"
                 )
 
             # Heating started/stopped
-            prev_heating = prev.currentInstPower > 0
-            curr_heating = status.currentInstPower > 0
+            prev_heating = prev.current_inst_power > 0
+            curr_heating = status.current_inst_power > 0
 
             if curr_heating and not prev_heating:
                 await self._event_emitter.emit("heating_started", status)
@@ -426,15 +514,17 @@ class MqttSubscriptionManager:
                 _logger.debug("Heating stopped")
 
             # Error detection
-            if status.errorCode and not prev.errorCode:
-                await self._event_emitter.emit("error_detected", status.errorCode, status)
-                _logger.info(f"Error detected: {status.errorCode}")
+            if status.error_code and not prev.error_code:
+                await self._event_emitter.emit(
+                    "error_detected", status.error_code, status
+                )
+                _logger.info(f"Error detected: {status.error_code}")
 
-            if not status.errorCode and prev.errorCode:
-                await self._event_emitter.emit("error_cleared", prev.errorCode)
-                _logger.info(f"Error cleared: {prev.errorCode}")
+            if not status.error_code and prev.error_code:
+                await self._event_emitter.emit("error_cleared", prev.error_code)
+                _logger.info(f"Error cleared: {prev.error_code}")
 
-        except Exception as e:
+        except (TypeError, AttributeError, RuntimeError) as e:
             _logger.error(f"Error detecting state changes: {e}", exc_info=True)
         finally:
             # Always update previous status
@@ -465,28 +555,36 @@ class MqttSubscriptionManager:
         Example::
 
             >>> def on_feature(feature: DeviceFeature):
-            ...     print(f"Serial: {feature.controllerSerialNumber}")
-            ...     print(f"FW Version: {feature.controllerSwVersion}")
-            ...     print(f"Temp Range: {feature.dhwTemperatureMin}-{feature.dhwTemperatureMax}°F")
+            ...     print(f"Serial: {feature.controller_serial_number}")
+            ...     print(f"FW Version: {feature.controller_sw_version}")
+            ...     print(f"Temp Range:
+            {feature.dhw_temperature_min}-{feature.dhw_temperature_max}°F")
             >>>
             >>> await mqtt_client.subscribe_device_feature(device, on_feature)
 
             >>> # Or use event emitter
-            >>> mqtt_client.on('feature_received', lambda f: print(f"FW: {f.controllerSwVersion}"))
-            >>> await mqtt_client.subscribe_device_feature(device, lambda f: None)
+            >>> mqtt_client.on('feature_received', lambda f: print(f"FW:
+            {f.controller_sw_version}"))
+            >>> await mqtt_client.subscribe_device_feature(device, lambda f:
+            None)
         """
 
-        def feature_message_handler(topic: str, message: dict[str, Any]) -> None:
+        def feature_message_handler(
+            topic: str, message: dict[str, Any]
+        ) -> None:
             """Parse feature messages and invoke user callback."""
             try:
                 # Log all messages received for debugging
-                _logger.debug(f"Feature handler received message on topic: {topic}")
+                _logger.debug(
+                    f"Feature handler received message on topic: {topic}"
+                )
                 _logger.debug(f"Message keys: {list(message.keys())}")
 
                 # Check if message contains feature data
                 if "response" not in message:
                     _logger.debug(
-                        "Message does not contain 'response' key, skipping. Keys: %s",
+                        "Message does not contain 'response' key, "
+                        "skipping. Keys: %s",
                         list(message.keys()),
                     )
                     return
@@ -496,13 +594,16 @@ class MqttSubscriptionManager:
 
                 if "feature" not in response:
                     _logger.debug(
-                        "Response does not contain 'feature' key, skipping. Keys: %s",
+                        "Response does not contain 'feature' key, "
+                        "skipping. Keys: %s",
                         list(response.keys()),
                     )
                     return
 
                 # Parse feature into DeviceFeature object
-                _logger.info(f"Parsing device feature message from topic: {topic}")
+                _logger.info(
+                    f"Parsing device feature message from topic: {topic}"
+                )
                 feature_data = response["feature"]
                 device_feature = DeviceFeature.from_dict(feature_data)
 
@@ -522,12 +623,18 @@ class MqttSubscriptionManager:
                     exc_info=True,
                 )
             except ValueError as e:
-                _logger.warning(f"Invalid value in feature message: {e}", exc_info=True)
-            except Exception as e:
-                _logger.error(f"Error parsing device feature: {e}", exc_info=True)
+                _logger.warning(
+                    f"Invalid value in feature message: {e}", exc_info=True
+                )
+            except (TypeError, AttributeError) as e:
+                _logger.error(
+                    f"Error parsing device feature: {e}", exc_info=True
+                )
 
         # Subscribe using the internal handler
-        return await self.subscribe_device(device=device, callback=feature_message_handler)
+        return await self.subscribe_device(
+            device=device, callback=feature_message_handler
+        )
 
     async def subscribe_energy_usage(
         self,
@@ -542,7 +649,8 @@ class MqttSubscriptionManager:
 
         Args:
             device: Device object
-            callback: Callback function that receives EnergyUsageResponse objects
+            callback: Callback function that receives EnergyUsageResponse
+            objects
 
         Returns:
             Subscription packet ID
@@ -550,24 +658,34 @@ class MqttSubscriptionManager:
         Example:
             >>> def on_energy_usage(energy: EnergyUsageResponse):
             ...     print(f"Total Usage: {energy.total.total_usage} Wh")
-            ...     print(f"Heat Pump: {energy.total.heat_pump_percentage:.1f}%")
-            ...     print(f"Electric: {energy.total.heat_element_percentage:.1f}%")
+            ...     print(f"Heat Pump:
+            {energy.total.heat_pump_percentage:.1f}%")
+            ...     print(f"Electric:
+            {energy.total.heat_element_percentage:.1f}%")
             >>>
-            >>> await mqtt_client.subscribe_energy_usage(device, on_energy_usage)
+            >>> await mqtt_client.subscribe_energy_usage(device,
+            on_energy_usage)
             >>> await mqtt_client.request_energy_usage(device, 2025, [9])
         """
-
         device_type = device.device_info.device_type
 
         def energy_message_handler(topic: str, message: dict[str, Any]) -> None:
-            """Internal handler to parse energy usage responses."""
+            """Parse and route energy usage responses to user callback.
+
+            Args:
+                topic: MQTT topic the message was received on
+                message: Parsed message dictionary
+            """
             try:
-                _logger.debug("Energy handler received message on topic: %s", topic)
+                _logger.debug(
+                    "Energy handler received message on topic: %s", topic
+                )
                 _logger.debug("Message keys: %s", list(message.keys()))
 
                 if "response" not in message:
                     _logger.debug(
-                        "Message does not contain 'response' key, skipping. Keys: %s",
+                        "Message does not contain 'response' key, "
+                        "skipping. Keys: %s",
                         list(message.keys()),
                     )
                     return
@@ -577,24 +695,38 @@ class MqttSubscriptionManager:
 
                 if "typeOfUsage" not in response_data:
                     _logger.debug(
-                        "Response does not contain 'typeOfUsage' key, skipping. Keys: %s",
+                        "Response does not contain 'typeOfUsage' key, "
+                        "skipping. Keys: %s",
                         list(response_data.keys()),
                     )
                     return
 
-                _logger.info("Parsing energy usage response from topic: %s", topic)
+                _logger.info(
+                    "Parsing energy usage response from topic: %s", topic
+                )
                 energy_response = EnergyUsageResponse.from_dict(response_data)
 
-                _logger.info("Invoking user callback with parsed EnergyUsageResponse")
+                _logger.info(
+                    "Invoking user callback with parsed EnergyUsageResponse"
+                )
                 callback(energy_response)
                 _logger.debug("User callback completed successfully")
 
             except KeyError as e:
-                _logger.warning("Failed to parse energy usage message - missing key: %s", e)
-            except Exception as e:
-                _logger.error("Error in energy usage message handler: %s", e, exc_info=True)
+                _logger.warning(
+                    "Failed to parse energy usage message - missing key: %s", e
+                )
+            except (TypeError, ValueError, AttributeError) as e:
+                _logger.error(
+                    "Error in energy usage message handler: %s",
+                    e,
+                    exc_info=True,
+                )
 
-        response_topic = f"cmd/{device_type}/{self._client_id}/res/energy-usage-daily-query/rd"
+        response_topic = (
+            f"cmd/{device_type}/{self._client_id}/res/"
+            f"energy-usage-daily-query/rd"
+        )
 
         return await self.subscribe(response_topic, energy_message_handler)
 
